/* This file is part of Mailfromd.
   Copyright (C) 2005-2011 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define MF_SOURCE_NAME  MF_SOURCE_SAVSRV
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <mailutils/mailutils.h>
#include <mailutils/daemon.h>

#include "mailfromd.h"

struct callout_command {
	const char *command;
	int argc;
	int (*handler) (FILE *, int, char **);
};

struct vrfy_queue {
	struct vrfy_queue *next;
	char *hostname;
	smtp_io_t io;
};

static struct vrfy_queue *head, *tail;

static void
verify(struct vrfy_queue *qp, FILE *fp)
{
	mf_status rc;

	if (qp->hostname) {
		mf_proctitle_format("callout server: <%s>@%s",
				    smtp_io_email(qp->io),
				    qp->hostname);
		rc = callout_strict(qp->io, qp->hostname);
		if (rc == mf_timeout)
			rc = mf_not_found;
		if (mf_resolved(rc))
			cache_insert2(smtp_io_email(qp->io),
				      qp->hostname, rc);		
	} else {
		mf_proctitle_format("callout server: %s",
				    smtp_io_email(qp->io));
		rc = callout_standard(qp->io);
		if (rc == mf_timeout)
			rc = mf_not_found;
		if (mf_resolved(rc))
			cache_insert(smtp_io_email(qp->io), rc);
	}
	if (fp)
		fprintf(fp, "* %s: %d %s\r\n",
			smtp_io_email(qp->io), rc, mf_status_str(rc));
}

static void
run_queue(FILE *fp)
{
	struct vrfy_queue *qp;

	for (qp = head; qp; ) {
		struct vrfy_queue *next = qp->next;
		verify(qp, fp);
		smtp_io_free(qp->io);
		free(qp->hostname);
		free(qp);
		qp = next;
	}
	head = tail = NULL;
}

static void
addq(int argc, char **argv)
{
	struct vrfy_queue *qp = xmalloc(sizeof(*qp));

	qp->io = smtp_io_create(argv[1],
				smtp_timeout_hard,
				NULL,
				NULL);
	smtp_io_setup_callout(qp->io, argv[1], argv[2], argv[3]);
	if (argv[4])
		qp->hostname = xstrdup(argv[4]);
	else
		qp->hostname = NULL;
	qp->next = NULL;
	if (tail)
		tail->next = qp;
	else
		head = qp;
	tail = qp;
}

int
cmd_vrfy(FILE *fp, int argc, char **argv)
{
	addq(argc, argv);
	fprintf(fp, "OK\r\n");
	return 0;
}

int
cmd_timeout(FILE *fp, int argc, char **argv)
{
	time_t to[SMTP_NUM_TIMEOUT];
	int i;
		
	for (i = 1; i < argc; i++) {
		char *p;
		unsigned long n = strtoul(argv[i], &p, 10);
		if (*p) {
			fprintf(fp, "NO syntax error in #%d\r\n", i);
			return 0;
		}
		to[i-1] = (time_t) n;
	}
	memcpy(smtp_timeout_hard, to, sizeof(to));
	fprintf(fp, "OK timeouts set\r\n");
	return 0;
}

int
cmd_run(FILE *fp, int argc, char **argv)
{
	run_queue(fp);
	fprintf(fp, "OK\r\n");
	return 0;
}

int
cmd_quit(FILE *fp, int argc, char **argv)
{
	fprintf(fp, "OK bye\r\n");
	return 1;
}

static struct callout_command callout_command_tab[] = {
	{ "VRFY", 3, cmd_vrfy },
	{ "SVRFY", 4, cmd_vrfy },
	{ "TIMEOUT", SMTP_NUM_TIMEOUT, cmd_timeout },
	{ "RUN", 0, cmd_run },
	{ "QUIT", 0, cmd_quit },
	{ NULL }
};

static struct callout_command *
find_command(const char *input)
{
	struct callout_command *cmd;

	for (cmd = callout_command_tab; cmd->command; cmd++) {
		if (strcasecmp(cmd->command, input) == 0)
			return cmd;
	}
	return NULL;
}

void
trimcrlf(char *buf)
{
	size_t len = strlen(buf);
	if (len >= 2 && buf[len-1] == '\n' && buf[len-2] == '\r')
		buf[len-2] = 0;
}


int
callout_session_server(const char *id, int fd,
		       struct sockaddr const *sa, socklen_t len,
		       void *server_data, void *srvman_data)
{
	FILE *fp = fdopen(fd, "w+");
	char *buf = NULL;
	size_t size = 0;

	mflog_reopen(id);
	mf_proctitle_format("callout server");
	setvbuf(fp, NULL, _IOLBF, 0);
	fprintf(fp, "OK mailfromd callout server ready\r\n");
	while (getline(&buf, &size, fp) >= 0) {
		int argc;
		char **argv;
		struct callout_command *cmd;
		int rc = 0;
		
		trimcrlf(buf);
		debug1(2,"C: %s", buf);

		if (mu_argcv_get(buf, "", NULL, &argc, &argv)) {
			fprintf(fp, "NO cannot parse line\r\n");
			continue;
		}

		cmd = find_command(argv[0]);
		if (!cmd)
			fprintf(fp, "NO unknown command\r\n");
		else if (argc - 1 != cmd->argc)
			fprintf(fp, "NO invalid arguments\r\n");
		else
			rc = cmd->handler(fp, argc, argv);
		mu_argcv_free(argc, argv);
		if (rc)
			break;
	}
	fclose(fp);
	free(buf);
	
	/* Run queued verifications */
	run_queue(NULL);
	return 0;
}
